@@ -2,19 +2,22 @@
 import abc
 import logging
 import os
-import psutil
 import socket
 import sys
+
+from contextlib import contextmanager
+from getpass import getpass, getuser
+from typing import Any, Dict, Iterator, List, Optional, TextIO, Tuple, TYPE_CHECKING, Union
+
+import psutil
 import yaml
 
-from getpass import getuser, getpass
-from contextlib import contextmanager
-from typing import Any, Dict, Iterator, List, Optional, TextIO, Tuple, TYPE_CHECKING, Union
 if TYPE_CHECKING:  # pragma: no cover
     from psycopg import Cursor
     from psycopg2 import cursor
 
 from . import psycopg
+from .collections import EMPTY_DICT
 from .config import Config
 from .exceptions import PatroniException
 from .log import PatroniLogger
@@ -22,7 +25,6 @@ from .postgresql.config import ConfigHandler, parse_dsn
 from .postgresql.misc import postgres_major_version_to_int
 from .utils import get_major_version, parse_bool, patch_config, read_stripped
 
-
 # Mapping between the libpq connection parameters and the environment variables.
 # This dict should be kept in sync with `patroni.utils._AUTH_ALLOWED_PARAMETERS`
 # (we use "username" in the Patroni config for some reason, other parameter names are the same).
@@ -71,8 +73,6 @@ class AbstractConfigGenerator(abc.ABC):
     :ivar config: dictionary used for the generated configuration storage.
     """
 
-    _HOSTNAME, _IP = get_address()
-
     def __init__(self, output_file: Optional[str]) -> None:
         """Set up the output file (if passed), helper vars and the minimal config structure.
 
@@ -91,14 +91,17 @@ class AbstractConfigGenerator(abc.ABC):
         :returns: dictionary with the values gathered from Patroni env, hopefully defined hostname and ip address
                   (otherwise set to :data:`~patroni.config_generator.NO_VALUE_MSG`), and some sane defaults.
         """
+        _HOSTNAME, _IP = get_address()
+
         template_config: Dict[str, Any] = {
             'scope': NO_VALUE_MSG,
-            'name': cls._HOSTNAME,
+            'name': _HOSTNAME,
             'restapi': {
-                'connect_address': cls._IP + ':8008',
-                'listen': cls._IP + ':8008'
+                'connect_address': _IP + ':8008',
+                'listen': _IP + ':8008'
             },
             'log': {
+                'type': PatroniLogger.DEFAULT_TYPE,
                 'level': PatroniLogger.DEFAULT_LEVEL,
                 'traceback_level': PatroniLogger.DEFAULT_TRACEBACK_LEVEL,
                 'format': PatroniLogger.DEFAULT_FORMAT,
@@ -106,8 +109,8 @@ class AbstractConfigGenerator(abc.ABC):
             },
             'postgresql': {
                 'data_dir': NO_VALUE_MSG,
-                'connect_address': cls._IP + ':5432',
-                'listen': cls._IP + ':5432',
+                'connect_address': _IP + ':5432',
+                'listen': _IP + ':5432',
                 'bin_dir': '',
                 'authentication': {
                     'superuser': {
@@ -125,6 +128,7 @@ class AbstractConfigGenerator(abc.ABC):
                 'noloadbalance': False,
                 'clonefrom': True,
                 'nosync': False,
+                'nostream': False,
             }
         }
 
@@ -242,7 +246,8 @@ class SampleConfigGenerator(AbstractConfigGenerator):
                   See :func:`~patroni.postgresql.misc.postgres_major_version_to_int` and
                   :func:`~patroni.utils.get_major_version`.
         """
-        postgres_bin = ((self.config.get('postgresql') or {}).get('bin_name') or {}).get('postgres', 'postgres')
+        postgres_bin = ((self.config.get('postgresql')
+                         or EMPTY_DICT).get('bin_name') or EMPTY_DICT).get('postgres', 'postgres')
         return postgres_major_version_to_int(get_major_version(self.config['postgresql'].get('bin_dir'), postgres_bin))
 
     def generate(self) -> None:
@@ -264,7 +269,8 @@ class SampleConfigGenerator(AbstractConfigGenerator):
         wal_level = 'hot_standby' if self.pg_major < 90600 else 'replica'
         self.config['bootstrap']['dcs']['postgresql']['parameters']['wal_level'] = wal_level
 
-        self.config['bootstrap']['dcs']['postgresql']['use_pg_rewind'] = True
+        self.config['bootstrap']['dcs']['postgresql']['use_pg_rewind'] = \
+            parse_bool(self.config['bootstrap']['dcs']['postgresql']['parameters']['wal_log_hints']) is True
         if self.pg_major >= 110000:
             self.config['postgresql']['authentication'].setdefault(
                 'rewind', {'username': 'rewind_user'}).setdefault('password', NO_VALUE_MSG)
@@ -395,8 +401,9 @@ class RunningClusterConfigGenerator(AbstractConfigGenerator):
             else:
                 self.config['bootstrap']['dcs']['postgresql']['parameters'][param] = value
 
+        connect_ip = self.config['postgresql']['connect_address'].rsplit(':')[0]
         connect_port = self.parsed_dsn.get('port', os.getenv('PGPORT', helper_dict['port']))
-        self.config['postgresql']['connect_address'] = f'{self._IP}:{connect_port}'
+        self.config['postgresql']['connect_address'] = f'{connect_ip}:{connect_port}'
         self.config['postgresql']['listen'] = f'{helper_dict["listen_addresses"]}:{helper_dict["port"]}'
 
     def _set_su_params(self) -> None:
@@ -409,8 +416,10 @@ class RunningClusterConfigGenerator(AbstractConfigGenerator):
             val = self.parsed_dsn.get(conn_param, os.getenv(env_var))
             if val:
                 su_params[conn_param] = val
-        patroni_env_su_username = ((self.config.get('authentication') or {}).get('superuser') or {}).get('username')
-        patroni_env_su_pwd = ((self.config.get('authentication') or {}).get('superuser') or {}).get('password')
+        patroni_env_su_username = ((self.config.get('authentication')
+                                    or EMPTY_DICT).get('superuser') or EMPTY_DICT).get('username')
+        patroni_env_su_pwd = ((self.config.get('authentication')
+                               or EMPTY_DICT).get('superuser') or EMPTY_DICT).get('password')
         # because we use "username" in the config for some reason
         su_params['username'] = su_params.pop('user', patroni_env_su_username) or getuser()
         su_params['password'] = su_params.get('password', patroni_env_su_pwd) or \
@@ -469,7 +478,7 @@ class RunningClusterConfigGenerator(AbstractConfigGenerator):
         with self._get_connection_cursor() as cur:
             self.pg_major = getattr(cur.connection, 'server_version', 0)
 
-            if not parse_bool(cur.connection.info.parameter_status('is_superuser')):
+            if not parse_bool(getattr(cur.connection, 'get_parameter_status')('is_superuser')):
                 raise PatroniException('The provided user does not have superuser privilege')
 
             self._set_pg_params(cur)
