@@ -1,14 +1,15 @@
 import abc
-from copy import deepcopy
 import logging
-import os
-import yaml
 
+from copy import deepcopy
 from typing import Any, Dict, Iterator, List, MutableMapping, Optional, Tuple, Type, Union
 
-from ..collections import CaseInsensitiveDict, CaseInsensitiveSet
+import yaml
+
+from ..collections import CaseInsensitiveDict
 from ..exceptions import PatroniException
 from ..utils import parse_bool, parse_int, parse_real
+from .available_parameters import get_validator_files, PathLikeObj
 
 logger = logging.getLogger(__name__)
 
@@ -258,10 +259,10 @@ class InvalidGucValidatorsFile(PatroniException):
     """Raised when reading or parsing of a YAML file faces an issue."""
 
 
-def _read_postgres_gucs_validators_file(file: str) -> Dict[str, Any]:
+def _read_postgres_gucs_validators_file(file: PathLikeObj) -> Dict[str, Any]:
     """Read an YAML file and return the corresponding Python object.
 
-    :param file: path to the file to be read. It is expected to be encoded with ``UTF-8``, and to be a YAML document.
+    :param file: path-like object to read from. It is expected to be encoded with ``UTF-8``, and to be a YAML document.
 
     :returns: the YAML content parsed into a Python object. If any issue is faced while reading/parsing the file, then
         return ``None``.
@@ -270,7 +271,7 @@ def _read_postgres_gucs_validators_file(file: str) -> Dict[str, Any]:
         :class:`InvalidGucValidatorsFile`: if faces an issue while reading or parsing *file*.
     """
     try:
-        with open(file, encoding='UTF-8') as stream:
+        with file.open(encoding='UTF-8') as stream:
             return yaml.safe_load(stream)
     except Exception as exc:
         raise InvalidGucValidatorsFile(
@@ -385,21 +386,7 @@ def _load_postgres_gucs_validators() -> None:
                     version_till: null
 
     """
-    conf_dir = os.path.join(
-        os.path.dirname(os.path.abspath(__file__)),
-        'available_parameters',
-    )
-    yaml_files: List[str] = []
-
-    for root, _, files in os.walk(conf_dir):
-        for file in sorted(files):
-            full_path = os.path.join(root, file)
-            if file.lower().endswith(('.yml', '.yaml')):
-                yaml_files.append(full_path)
-            else:
-                logger.info('Ignored a non-YAML file found under `available_parameters` directory: `%s`.', full_path)
-
-    for file in yaml_files:
+    for file in get_validator_files():
         try:
             config: Dict[str, Any] = _read_postgres_gucs_validators_file(file)
         except InvalidGucValidatorsFile as exc:
@@ -425,9 +412,8 @@ _load_postgres_gucs_validators()
 
 
 def _transform_parameter_value(validators: MutableMapping[str, Tuple[_Transformable, ...]],
-                               version: int, name: str, value: Any,
-                               available_gucs: CaseInsensitiveSet) -> Optional[Any]:
-    """Validate *value* of GUC *name* for Postgres *version* using defined *validators* and *available_gucs*.
+                               version: int, name: str, value: Any) -> Optional[Any]:
+    """Validate *value* of GUC *name* for Postgres *version* using defined *validators*.
 
     :param validators: a dictionary of all GUCs across all Postgres versions. Each key is the name of a Postgres GUC,
         and the corresponding value is a variable length tuple of :class:`_Transformable`. Each item is a validation
@@ -436,8 +422,6 @@ def _transform_parameter_value(validators: MutableMapping[str, Tuple[_Transforma
     :param version: Postgres version to validate the GUC against.
     :param name: name of the Postgres GUC.
     :param value: value of the Postgres GUC.
-    :param available_gucs: a set of all GUCs available in Postgres *version*. Each item is the name of a Postgres
-        GUC. Used for a couple purposes:
 
         * Disallow writing GUCs to ``postgresql.conf`` (or ``recovery.conf``) that does not exist in Postgres *version*;
         * Avoid ignoring GUC *name* if it does not have a validator in *validators*, but is a valid GUC in Postgres
@@ -445,35 +429,23 @@ def _transform_parameter_value(validators: MutableMapping[str, Tuple[_Transforma
 
     :returns: the return value may be one among:
 
-        * *value* transformed to the expected format for GUC *name* in Postgres *version*, if *name* is present
-          in *available_gucs* and has a validator in *validators* for the corresponding Postgres *version*; or
-        * The own *value* if *name* is present in *available_gucs* but not in *validators*; or
-        * ``None`` if *name* is not present in *available_gucs*.
+        * *value* transformed to the expected format for GUC *name* in Postgres *version*, if *name* has a validator
+          in *validators* for the corresponding Postgres *version*; or
+        * ``None`` if *name* does not have a validator in *validators*.
     """
-    if name in available_gucs:
-        for validator in validators.get(name, ()) or ():
-            if version >= validator.version_from and\
-                    (validator.version_till is None or version < validator.version_till):
-                return validator.transform(name, value)
-        # Ideally we should have a validator in *validators*. However, if none is available, we will not discard a
-        # setting that exists in Postgres *version*, but rather allow the value with no validation.
-        return value
+    for validator in validators.get(name, ()) or ():
+        if version >= validator.version_from and\
+                (validator.version_till is None or version < validator.version_till):
+            return validator.transform(name, value)
     logger.warning('Removing unexpected parameter=%s value=%s from the config', name, value)
 
 
-def transform_postgresql_parameter_value(version: int, name: str, value: Any,
-                                         available_gucs: CaseInsensitiveSet) -> Optional[Any]:
-    """Validate *value* of GUC *name* for Postgres *version* using ``parameters`` and *available_gucs*.
+def transform_postgresql_parameter_value(version: int, name: str, value: Any) -> Optional[Any]:
+    """Validate *value* of GUC *name* for Postgres *version* using ``parameters``.
 
     :param version: Postgres version to validate the GUC against.
     :param name: name of the Postgres GUC.
     :param value: value of the Postgres GUC.
-    :param available_gucs: a set of all GUCs available in Postgres *version*. Each item is the name of a Postgres
-        GUC. Used for a couple purposes:
-
-            * Disallow writing GUCs to ``postgresql.conf`` that does not exist in Postgres *version*;
-            * Avoid ignoring GUC *name* if it does not have a validator in ``parameters``, but is a valid GUC in
-                Postgres *version*.
 
     :returns: The return value may be one among:
 
@@ -488,32 +460,17 @@ def transform_postgresql_parameter_value(version: int, name: str, value: Any,
         return value
     if name in recovery_parameters:
         return None
-    return _transform_parameter_value(parameters, version, name, value, available_gucs)
+    return _transform_parameter_value(parameters, version, name, value)
 
 
-def transform_recovery_parameter_value(version: int, name: str, value: Any,
-                                       available_gucs: CaseInsensitiveSet) -> Optional[Any]:
-    """Validate *value* of GUC *name* for Postgres *version* using ``recovery_parameters`` and *available_gucs*.
+def transform_recovery_parameter_value(version: int, name: str, value: Any) -> Optional[Any]:
+    """Validate *value* of GUC *name* for Postgres *version* using ``recovery_parameters``.
 
     :param version: Postgres version to validate the recovery GUC against.
     :param name: name of the Postgres recovery GUC.
     :param value: value of the Postgres recovery GUC.
-    :param available_gucs: a set of all GUCs available in Postgres *version*. Each item is the name of a Postgres
-        GUC. Used for a couple purposes:
-
-            * Disallow writing GUCs to ``recovery.conf`` (or ``postgresql.conf`` depending on *version*), that does not
-                exist in Postgres *version*;
-            * Avoid ignoring recovery GUC *name* if it does not have a validator in ``recovery_parameters``, but is a
-                valid GUC in Postgres *version*.
 
     :returns: *value* transformed to the expected format for recovery GUC *name* in Postgres *version* using validators
         defined in ``recovery_parameters``. It can also return ``None``. See :func:`_transform_parameter_value`.
     """
-    # Recovery settings are not present in ``postgres --describe-config`` output of Postgres <= 11. In that case we
-    # just pass down the list of settings defined in Patroni validators so :func:`_transform_parameter_value` will not
-    # discard the recovery GUCs when running Postgres <= 11.
-    # NOTE: At the moment this change was done Postgres 11 was almost EOL, and had been likely extensively used with
-    # Patroni, so we should be able to rely solely on Patroni validators as the source of truth.
-    return _transform_parameter_value(
-        recovery_parameters, version, name, value,
-        available_gucs if version >= 120000 else CaseInsensitiveSet(recovery_parameters.keys()))
+    return _transform_parameter_value(recovery_parameters, version, name, value)
